from typing import Tuple

import numpy as np

from mu_map.file.dicom import DICOM
from mu_map.file.interfile import (
    Interfile,
    parse_interfile_header_str,
    TEMPLATE_HEADER_IMAGE,
)

"""
Module for the conversion between the different formats.
"""


def parse_position(
    dcm: DICOM, default_orientation: str = "feet_in", default_rotation: str = "supine"
) -> Tuple[str, str]:
    """
    Parse the patient position from the DICOM patient position
    field into the orientation and rotation as used by the Interfile
    format.

    :param position: the value of the patient position field in the DICOM header
    :return: a tuple of patient orientation and rotation
    """
    orientation = default_orientation
    rotation = default_rotation

    try:
        position = dcm.PatientPosition

        _orienation = position[:2]
        if _orienation == "FF":
            orientation = "feet_in"
        elif _orienation == "HF":
            orientation = "head_in"

        _rotation = position[2]
        if _orienation == "P":
            rotation = "prone"
        elif _rotation == "S":
            rotation = "supine"
    except:
        pass
    return orientation, rotation


def dicom_to_interfile(dcm: DICOM, image: np.ndarray) -> Interfile:
    """
    Convert a DICOM image to Interfile format.

    :param dcm: the DICOM image
    :param image: the image as a numpy array
    :return: an Interfile as a combination of header and image
    """

    header = TEMPLATE_HEADER_IMAGE.replace("{ROWS}", str(dcm.Rows))
    header = header.replace("{SPACING_X}", f"{dcm.PixelSpacing[0]:.4f}")
    header = header.replace("{COLUMNS}", str(dcm.Columns))
    header = header.replace("{SPACING_Y}", f"{dcm.PixelSpacing[1]:.4f}")
    header = header.replace("{SLICES}", str(dcm.NumberOfSlices))
    header = header.replace("{SPACING_Z}", f"{dcm.SliceThickness:.4f}")
    header = header.replace("{OFFSET_X}", str(-0.5 * dcm.Rows * dcm.PixelSpacing[0]))
    header = header.replace("{OFFSET_Y}", str(-0.5 * dcm.Columns * dcm.PixelSpacing[1]))

    orientation, rotation = parse_position(dcm)
    header = header.replace("{PATIENT_ORIENTATION}", orientation)
    header = header.replace("{PATIENT_ROTATION}", rotation)

    header = parse_interfile_header_str(header)
    return Interfile(header, image.astype(np.float32))


if __name__ == "__main__":
    import argparse
    import os

    import pydicom

    from mu_map.file.dicom import change_uid, load_dcm, update_dcm, EXTENSION_DICOM
    from mu_map.file.interfile import write_interfile, load_interfile_img

    def main_interfile(args):
        args.out = args.file if args.out is None else args.out

        dcm, image = load_dcm(args.file)
        interfile = dicom_to_interfile(dcm, image)
        write_interfile(args.out, interfile)

    def main_dicom(args):
        args.out = args.file if args.out is None else args.out
        args.out = os.path.splitext(args.out)[0]
        args.out = f"{args.out}{EXTENSION_DICOM}"

        image = load_interfile_img(args.file)
        reference = pydicom.dcmread(args.reference)
        dcm = update_dcm(reference.copy(), image)
        dcm = change_uid(dcm)

        if args.description:
            dcm.SeriesDescription = args.description

        pydicom.dcmwrite(args.out, dcm)

    parser = argparse.ArgumentParser(
        description="convert images between different formats",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "file",
        type=str,
        help="the file to be converted (a DICOM image or an Interfile header)",
    )
    parser.add_argument(
        "--out",
        type=str,
        help="the filename for the output file. If not specified it defaults to the input name with different extensions.",
    )
    subparsers = parser.add_subparsers(help="file to convert to")

    parser_dcm = subparsers.add_parser("dicom", help="convert to a DICOM file")
    parser_dcm.add_argument(
        "--reference",
        type=str,
        required=True,
        help="reference DICOM file to use for the header",
    )
    parser_dcm.add_argument(
        "--description", type=str, help="update the series description"
    )
    parser_dcm.set_defaults(func=main_dicom)

    parser_inter = subparsers.add_parser("interfile", help="convert to an Interfile")
    parser_inter.set_defaults(func=main_interfile)

    args = parser.parse_args()
    args.func(args)
