from functools import reduce
from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn


class Conv(nn.Sequential):
    """
    A wrapper around a 3D convolutional layer that also contains batch normalization, a ReLU activation function and a max pooling.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
    ):
        """
        Create a convolutional layer with batch normalization, a ReLU activation function and max pooling.

        Parameters
        ----------
        in_channels: int
            number of channels received as input
        out_channels: int
            number of filters and consequently channels in the output
        """
        super().__init__()

        self.append(
            nn.Conv3d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=1,
                padding="same",
            )
        )
        self.append(nn.BatchNorm3d(num_features=out_channels))
        self.append(nn.ReLU(inplace=True))
        self.append(nn.MaxPool3d(kernel_size=2, stride=2))


class Discriminator(nn.Module):
    """
    Create the discriminator as specified by Shi et al. (2020).
    It consists of three convolutional layers with max pooling, followed by three fully connected layers.
    """

    def __init__(
        self,
        in_channels: int,
        input_size: Union[int, Tuple[int, int, int]],
        conv_features: List[int] = [32, 64, 128],
        fc_features: List[int] = [512, 128],
    ):
        """
        Create the discriminator.

        Parameters
        ----------
        in_channels: int
            number channels received as an input
        input_size: int or tuple of int
            shape of the inputs either as an int (if equal in all dimensions) or as a tuple
            this is required to compute the number of features for the first fully connected layer
        conv_features: list of int
            each number represents the number of filter in a convolutional layer
        fc_features: list of int
            each number represents the number of features in a fully connected layer
            note that there will always be an additional fully connected layer between these features and the convolutions
        """
        super().__init__()

        self.in_channels = in_channels
        self.input_size = input_size if type(input_size) is tuple else (input_size,) * 3
        self.conv_features = conv_features
        self.fc_features = fc_features

        conv_layers = []
        for in_channels, out_channels in zip(
            [in_channels, *conv_features[:-1]], conv_features
        ):
            conv_layers.append(Conv(in_channels, out_channels))
        self.conv = nn.Sequential(*conv_layers)

        # input is halved by each convolutional layer
        self.fc_input_size = map(lambda x: x // 2 ** (len(conv_features)), input_size)
        self.fc_input_size = reduce(lambda x, y: x * y, self.fc_input_size)

        fc_layers = []
        fc_layers.append(
            nn.Linear(
                in_features=self.fc_input_size * conv_features[-1],
                out_features=fc_features[0],
            )
        )
        fc_layers.append(nn.ReLU(inplace=True))
        for in_channels, out_channels in zip([*fc_features[:-1]], fc_features[1:]):
            fc_layers.append(
                nn.Linear(in_features=in_channels, out_features=out_channels)
            )
            fc_layers.append(nn.ReLU(inplace=True))
        fc_layers.append(nn.Linear(in_features=fc_features[-1], out_features=1))
        self.fc = nn.Sequential(*fc_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x

    def get_output_shape(self, input_shape: Tuple[int]) -> Tuple[int]:
        """
        Get the output shape of the discriminator with respect to an input shape.

        Parameters
        ----------
        input_shape: Tuple[int]
            the shape for which the output is computed which needs to have at least 4 dimension

        Returns
        -------
        Tuple[int]
            the according output shape
        """
        assert len(input_shape) > 3, "Input must contain at least 4 dimensions"
        return (*input_shape[:-4], 1)


class PatchDiscriminator(nn.Sequential):
    """
    Patch discriminator (PatchGAN) used by Isola et al. 2017 for their image-to-image
    translation GAN.
    """

    def __init__(self, in_channels: int = 2, features: List[int] = [64, 128, 256, 512]):
        """
        Create a new patch discriminator.

        Parameters
        ----------
        in_channels: int
            the number of channels of inputs for the model
        features: list of int
            the number of features in each convolutional layer of the model
            note that each feature introduces a new layer that halves the input size
        """
        super().__init__()
        self.features = features
        self.in_channels = in_channels

        self.append(self.create_conv(in_channels, features[0]))
        self.append(nn.LeakyReLU(inplace=True, negative_slope=0.2))
        for in_channels, out_channels in zip(features[:-1], features[1:]):
            self.append(self.create_conv(in_channels, out_channels))
            self.append(nn.BatchNorm3d(num_features=out_channels))
            self.append(nn.LeakyReLU(inplace=True, negative_slope=0.2))
        self.append(self.create_conv(features[-1], 1))

    def create_conv(self, in_channels: int, out_channels: int) -> nn.Conv3d:
        """
        Create a typical convolution used by the patch discriminator (kernel_size=4, stride=2, padding=1).

        Parameters
        ----------
        in_channels: int
            the number of input channels for the convolution
        out_channels: int
            the number of output channels (number of filters) of the convolution

        Returns
        -------
        nn.Conv3d
        """
        return nn.Conv3d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=4,
            stride=2,
            padding=1,
        )

    def get_output_shape(self, input_shape: Tuple[int]) -> Tuple[int]:
        """
        Get the output shape of the discriminator with respect to an input shape.

        Parameters
        ----------
        input_shape: Tuple[int]
            the shape for which the output is computed which needs to have at least 4 dimension

        Returns
        -------
        Tuple[int]
            the according output shape
        """
        assert len(input_shape) > 3, "Input must contain at least 4 dimensions"

        n_convs = len(self.features) + 1  # each convolution halves the input dimensions
        return (
            *input_shape[:-4],
            1,
            *map(lambda x: x // (n_convs**2), input_shape[-3:]),
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="print and test different discriminators",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--type",
        choices=["class", "patch"],
        default="class",
        help="test either a classification or a PatchGAN discriminator",
    )
    parser.add_argument(
        "--conv_features",
        type=int,
        nargs="+",
        default=[32, 64, 128],
        help="the number of features for each convolutional layer",
    )
    parser.add_argument(
        "--fc_features",
        type=int,
        nargs="+",
        default=[512, 128],
        help="the number of features for each fully connected layer",
    )
    parser.add_argument(
        "--input_size",
        type=int,
        nargs=3,
        default=[32, 32, 32],
        help="shape of inputs for the discriminator",
    )
    parser.add_argument(
        "--in_channels", type=int, default=2, help="number of channels for inputs"
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=8,
        help="batch size of inputs for a test computation",
    )
    args = parser.parse_args()

    if args.type == "class":
        net = Discriminator(
            in_channels=args.in_channels,
            input_size=args.input_size,
            conv_features=args.conv_features,
            fc_features=args.fc_features,
        )
    elif args.type == "patch":
        net = PatchDiscriminator(
            in_channels=args.in_channels, features=args.conv_features
        )
    print(net)

    _inputs = torch.rand((args.batch_size, args.in_channels, *args.input_size))
    _outputs = net(_inputs)

    _targets = torch.full(_outputs.shape, 1.0)
    criterion = torch.nn.MSELoss()
    loss = criterion(_outputs, _targets)
    print(loss.item())

    print(f"Transform {_inputs.shape} to {_outputs.shape}")
    assert _outputs.shape == net.get_output_shape(_inputs.shape)
