from typing import List, Optional

import matplotlib as mlp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from mu_map.polar_map.prepare import headers

SIZE_DEFAULT = 12
plt.rc("font", family="Arial")  # controls default font
plt.rc("font", weight="normal")  # controls default font
plt.rc("font", size=SIZE_DEFAULT)  # controls default text sizes
plt.rc("axes", titlesize=16)  # fontsize of the axes title


# COLORS=["#a6cee3", "#1f78b4", "#b2df8a"]
COLORS=["#66c2a5", "#fc8d62", "#8da0cb"]
COLORS=COLORS[::-1]

def bland_altman(
    data1: np.ndarray, data2: np.ndarray, ax: Optional[mlp.axes.Axes] = None
):
    ax = plt.subplot() if ax is None else ax

    mean = np.mean([data1, data2], axis=0)
    diff = data1 - data2  # Difference between data1 and data2
    md = np.mean(diff)  # Mean of the difference
    sd = np.std(diff, axis=0)  # Standard deviation of the difference

    ax.axhline(md, color="#fc8d59", linestyle="-")
    for x in [-1.96, 1.96]:
        ax.axhline(md + x * sd, color="black", linestyle="--", alpha=0.8)
    ax.scatter(mean, diff, color="#91bfdb", s=25, alpha=0.7, edgecolors="black", linewidths=0.3)

    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)

    ax.grid(axis="both", alpha=0.5, linestyle="dotted")



data = pd.read_csv("data/polar_maps/perfusion.csv")

baseline = data[data[headers.ac] & (data[headers.type] == "symbia")]

correction_none = data[~data[headers.ac]]
correction_syn = data[data[headers.ac] & (data[headers.type] == "synthetic")]
correction_ct = data[data[headers.ac] & (data[headers.type] == "ct")]

keys_segments = [f"segment_{i}" for i in range(1, 18)]

_baseline = baseline[keys_segments].values
_correction_none = correction_none[keys_segments].values
_correction_syn = correction_syn[keys_segments].values
_correction_ct = correction_ct[keys_segments].values

fig, axs = plt.subplots(1, 2, figsize=(12, 6))
bland_altman(_correction_none.flatten(), _correction_ct.flatten(), axs[0])
bland_altman(_correction_syn.flatten(), _correction_ct.flatten(), axs[1])
axs[0].set_title("No Attenuation Correction")
axs[1].set_title("Synthetic Attenuation Correction")

# fig, axs = plt.subplots(1, 3, figsize=(18, 6))
# bland_altman(_baseline.flatten(), _correction_none.flatten(), ax=axs[0])
# bland_altman(_baseline.flatten(), _correction_syn.flatten(), ax=axs[1])
# bland_altman(_baseline.flatten(), _correction_ct.flatten(), ax=axs[2])


def normalize_axes(axes: List[mlp.axes.Axes]):
    x_min = min(map(lambda ax: ax.get_xlim()[0], axes))
    x_max = max(map(lambda ax: ax.get_xlim()[1], axes))
    y_min = min(map(lambda ax: ax.get_ylim()[0], axes))
    y_max = max(map(lambda ax: ax.get_ylim()[1], axes))
    for ax in axs:
        ax.set_xlim((x_min, x_max))
        ax.set_ylim((y_min, y_max))


normalize_axes(axs)
# plt.hist(_correction_syn.flatten() - _correction_ct.flatten())
# plt.hist(_correction_none.flatten() - _correction_ct.flatten())
plt.show()
exit(0)


def absolute_percent_error(prediction: np.ndarray, target: np.ndarray) -> float:
    mean_p = prediction.mean(axis=0)
    mean_t = target.mean(axis=0)
    diff = np.absolute(mean_p - mean_t)
    return 100.0 * diff / mean_t


def apc(prediction: np.ndarray, target: np.ndarray) -> float:
    return absolute_percent_error(prediction, target)


combinations = [
    ("NoAC to AC", _correction_none, _baseline),
    ("SYN to AC", _correction_syn, _baseline),
    ("CT to AC", _correction_ct, _baseline),
    ("NoAC to CT", _correction_none, _correction_ct),
    ("SYN to CT", _correction_syn, _correction_ct),
]

for label, prediction, target in combinations:
    apcs = absolute_percent_error(prediction, target)
    _mean = apcs.mean()
    _std = apcs.std()
    print(f"APE - {label:>12}: {_mean:.3f}±{_std:.3f}")


n_plots = 5
plot_width = 4
fig, axs = plt.subplots(1, n_plots, figsize=(n_plots * plot_width + 1, plot_width + 1))
fig.suptitle("Correlation of Segment Perfusion")


def jitter(array: np.ndarray, scale=1):
    return array + (np.random.random(array.shape) * scale - 0.5 * scale)


axs[0].scatter(
    jitter(_baseline.flatten()),
    jitter(_correction_none.flatten()),
    c="#67a9cf",
    s=20,
    alpha=0.6,
    edgecolor="black",
)
axs[1].scatter(
    jitter(_baseline.flatten()),
    jitter(_correction_syn.flatten()),
    c="#67a9cf",
    s=20,
    alpha=0.6,
    edgecolor="black",
)
axs[2].scatter(
    jitter(_baseline.flatten()),
    jitter(_correction_ct.flatten()),
    c="#67a9cf",
    s=20,
    alpha=0.6,
    edgecolor="black",
)

axs[3].scatter(
    jitter(_correction_ct.flatten()),
    jitter(_correction_none.flatten()),
    c="#67a9cf",
    s=20,
    alpha=0.6,
    edgecolor="black",
)
axs[4].scatter(
    jitter(_correction_ct.flatten()),
    jitter(_correction_syn.flatten()),
    c="#67a9cf",
    s=20,
    alpha=0.6,
    edgecolor="black",
)

axs[0].set_xlabel("Perfusion AC")
axs[1].set_xlabel("Perfusion AC")
axs[2].set_xlabel("Perfusion AC")
axs[3].set_xlabel("Perfusion AC CT")
axs[4].set_xlabel("Perfusion AC CT")

axs[0].set_ylabel("Perfusion NoAC")
axs[1].set_ylabel("Perfusion AC SYN")
axs[2].set_ylabel("Perfusion AC CT")
axs[3].set_ylabel("Perfusion NoAC")
axs[4].set_ylabel("Perfusion AC SYN")

axs[0].set_title("  NoAC vs. AC")
axs[1].set_title("AC SYN vs. AC")
axs[2].set_title(" AC CT vs. AC")
axs[3].set_title("  NoAC vs. AC CT")
axs[4].set_title("AC SYN vs. AC CT")

for ax in axs:
    ax.set_xlim((30, 100))
    ax.set_ylim((30, 100))
    ax.plot((30, 100), (30, 100), color="#ef8a62")

plt.tight_layout()
plt.show()
